#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <GL/glew.h>
#include <GL/glu.h>

#include <SDL.h>
#include <SDL_opengl.h>

#include "utils_graphics.h"


#define MIN(a, b)           ((a) < (b)) ? (a) : (b);
#define MAX(a, b)           ((a) > (b)) ? (a) : (b);
#define SCREEN_WIDTH        640
#define SCREEN_HEIGHT       480

typedef struct {
    float real;
    float imaginary;
} t_complex;


SDL_Window* gWindow = NULL;
SDL_GLContext gContext;

GLuint vertex_shader;
GLuint fragment_shader;
GLuint shader_program;
GLuint a_point_position_loc;
GLuint a_point_color_loc;

GLuint vertex_array;
GLuint vertex_buffer;

float projection_matrix[16];
float modelview_matrix[16];
float aux_matrix1[16];
float aux_matrix2[16];

GLfloat ratioX, ratioY;
GLfloat minX = -0.6f;
GLfloat maxX = 0.2f;
GLfloat minY = 0.0f;
GLfloat maxY = 0.8f;
GLfloat colors[121*3];

GLfloat z_coord = 0.0f;
GLfloat min_x_coord = -4.0f;
GLfloat max_x_coord = 4.0f;
GLfloat min_y_coord = -2.9f;
GLfloat max_y_coord = 2.9f;

GLfloat x_step = 0.01f;
GLfloat y_step = 0.01f;

unsigned long size = 0;
unsigned long STRIDE = 0;



int init();
int initGL();
int check_gl_error();
void cleanup();
int set_shaders();
int init_buffers();
void update();
void render();

char *text_file_read(char *fname);
void print_shader_info_log(GLuint obj);
void print_program_info_log(GLuint obj);



void HSV2RGB(float hue, float sat, float val, float *r, float *g, float *b) {
    int i;
    float f, p, q, t;

    if (val == 0) {
        *r = 0.0f; *g = 0.0f; *b = 0.0f;
        return;
    } else {
        hue /= 60.0f;
        i = round(hue);
        f = hue-i;
        p = val * (1 - sat);
        q = val * (1 - (sat * f));
        t = val * (1 - (sat * (1 - f)));

        if (i == 0) {
            *r = val; *g = t; *b = p;
            return;
        } else if (i == 1) {
            *r = q; *g = val; *b = p;
            return;
        } else if (i == 2) {
            *r = p; *g = val; *b = t;
            return;
        } else if (i == 3) {
            *r = p; *g = q; *b = val;
            return;
        } else if (i == 4) {
            *r = t; *g = p; *b = val;
            return;
        } else if (i == 5) {
            *r = val; *g = p; *b = q;
            return;
        }
    }
}

//calculate squared modus of given complex c
float complex_mod_sq(t_complex complex) {
    return complex.real*complex.real + complex.imaginary*complex.imaginary;
}

//function z[0]=0
t_complex f() {
    t_complex complex;

    complex.real = 0.0f;
    complex.imaginary = 0.0f;

    return complex;
}

//function z[n+1] = z[n]^2 + p
t_complex g(t_complex z, t_complex p) {
    t_complex complex;

    complex.real = z.real*z.real - z.imaginary*z.imaginary + p.real;
    complex.imaginary = 2*z.real*z.imaginary + p.imaginary;

    return complex;
}

//value is inside set in the returned level
long level_set(t_complex p) {
    t_complex z;
    long iteration;

    iteration = 0;
    z = f();

    do {
        z = g(z, p);
        iteration++;
    } while (complex_mod_sq(z) < 4 && iteration < 120);

    return iteration;
}

void initialize_colors() {
    int level;
    float r, g, b;

    for (level = 0; level < 120; level++) {
        HSV2RGB(2.1*level, 0.85, 0.6, &r, &g, &b);
        colors[level*3] = r;
        colors[level*3+1] = g;
        colors[level*3+2] = b;
    }
    colors[120*3] = 0;
    colors[120*3+1] = 0;
    colors[120*3+2] = 0;
}

int init() {
    // Initialize SDL
    if (SDL_Init(SDL_INIT_VIDEO) < 0) {
        printf("Błąd podczas inicjalizacji silnika SDL2: %s\n", SDL_GetError());
        return 0;
    }

    SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, 1);
    SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
    SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 24);

    gWindow = SDL_CreateWindow("Fraktal Mandelbrota", SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED,
                                SCREEN_WIDTH, SCREEN_HEIGHT, SDL_WINDOW_OPENGL | SDL_WINDOW_SHOWN);
    if (gWindow == NULL) {
        printf("Błąd podczas tworzenia okna: %s\n", SDL_GetError());
        return 0;
    }

    gContext = SDL_GL_CreateContext(gWindow);
    if (gContext == NULL) {
        printf("Błąd podczas tworzenia kontekstu OpenGL: %s\n", SDL_GetError());
        return 0;
    }

    if (SDL_GL_SetSwapInterval(1) < 0) {
        printf( "Błąd podczas ustawiania VSync: %s\n", SDL_GetError());
    }

    initialize_colors();

    if (!initGL()) {
        printf("Błąd podczas inicjalizacji OpenGL!\n");
        return 0;
    }

    return 1;
}

int initGL() {
    GLenum err = GL_NO_ERROR;
    GLfloat aspect = 1.0f;
    unsigned int width = SCREEN_WIDTH;
    unsigned int height = SCREEN_HEIGHT;

    glewExperimental = GL_TRUE;
    err = glewInit();
    while (check_gl_error() == GL_INVALID_ENUM) {};
    if (err != GLEW_OK) {
        printf("glewInit failed: %s\n", glewGetErrorString(err));
        return 0;
    }

    glClearColor(0.f, 0.f, 0.f, 1.f);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_VERTEX_PROGRAM_POINT_SIZE);

    glViewport(0, 0, width, height);

    if (height > 0) {
        aspect = (float)width / (float)height;
    }

    get_perspective_matrix(projection_matrix, 45.0, aspect, 0.1, 100.0);

    if (!set_shaders()) {
        cleanup();
        return -1;
    }
    if (!init_buffers()) {
        cleanup();
        return -1;
    }

    return 1;
}


int set_shaders() {
    char *shader_source = NULL;
    GLint status;

    shader_source = text_file_read("shader_vertex.vshader");
    if (!shader_source) {
        printf("Nie można odczytać shadera wierzchołków\n");
        return 0;
    }
    //printf("%s\n---\n", shader_source);
    vertex_shader = glCreateShader(GL_VERTEX_SHADER);
    //printf("vertex_shader: %d\n", vertex_shader);
    glShaderSource(vertex_shader,  1, (const char* const*) &shader_source, NULL);
    free(shader_source);
    glCompileShader(vertex_shader);
    glGetShaderiv(vertex_shader, GL_COMPILE_STATUS, &status);
    if (status != GL_TRUE) {
        printf("Niepoprawna kompilacja shadera wierzchołków\n");
        print_shader_info_log(vertex_shader);
        return 0;
    }

    shader_source = text_file_read("shader_fragment.fshader");
    if (!shader_source) {
        printf("Nie można odczytać shadera fragmentów\n");
        return 0;
    }
    //printf("%s\n---\n", shader_source);
    fragment_shader = glCreateShader(GL_FRAGMENT_SHADER);
    //printf("fragment_shader: %d\n", fragment_shader);
    glShaderSource(fragment_shader,  1, (const char* const*) &shader_source, NULL);
    free(shader_source);
    glCompileShader(fragment_shader);
    glGetShaderiv(fragment_shader, GL_COMPILE_STATUS, &status);
    if (status != GL_TRUE) {
        printf("Niepoprawna kompilacja shadera fragmentów\n");
        print_shader_info_log(vertex_shader);
        return 0;
    }

    shader_program = glCreateProgram();
    glAttachShader(shader_program, vertex_shader);
    glAttachShader(shader_program, fragment_shader);
    glLinkProgram(shader_program);
    glGetProgramiv(shader_program, GL_LINK_STATUS, &status);
    if (status == GL_FALSE) {
        printf("Niepoprawna konsolidacja programu\n");
        print_program_info_log(shader_program);
        return 0;
    }
    glValidateProgram(shader_program);
    glGetProgramiv(shader_program, GL_VALIDATE_STATUS, &status);
    if (status == GL_FALSE) {
        printf("Niepoprawna walidacja programu\n");
        print_program_info_log(shader_program);
        return 0;
    }

    glUseProgram(shader_program);

    return 1;
}


int init_buffers() {
    GLfloat i, j;
    unsigned long index;
    long level;
    GLfloat *data_array;
    t_complex p;
    p.real = 0.0f;
    p.imaginary = 0.0f;

    ratioX = (maxX - minX) / 2.0;
    ratioY = (maxY - minY) / 2.0;

    size = 0;
    for (i = min_x_coord; i < max_x_coord; i += x_step) {
        for (j = min_y_coord; j < max_y_coord; j += y_step) {
            size++;
        }
    }

    data_array = malloc(size * sizeof(GLfloat) * (3+3));
    if (!data_array) {
        return 0;
    }

    index = 0;
    for (i = min_x_coord; i < max_x_coord; i += x_step) {
        p.imaginary = i*ratioY + minY;
        for (j = min_y_coord; j < max_y_coord; j += y_step) {
            p.real = j*ratioX + minX;
            level = level_set(p);
            data_array[(3+3)*index] = i;
            data_array[(3+3)*index+1] = j;
            data_array[(3+3)*index+2] = z_coord;

            data_array[(3+3)*index+3] = colors[level*3];
            data_array[(3+3)*index+4] = colors[level*3 + 1];
            data_array[(3+3)*index+5] = colors[level*3 + 2];

            index++;
        }
    }

    STRIDE = (3+3) * sizeof(GLfloat);
    /*
    for (index = 0; index < size; index++) {
        printf("%.2f  %.2f  %.2f;    %.2f  %.2f %.2f\n",
            data_array[6*index], data_array[6*index+1], data_array[6*index+2],
            data_array[6*index+3], data_array[6*index+4], data_array[6*index+5]);
    }
    printf("---\n");
    for (index = 0; index < size * 6; index++) {
        printf("%.2f  ", data_array[index]);
        if (index % 6 == 5) {
            printf("\n");
        }
    }
    */

    glGenVertexArrays(1, &vertex_array);
    glBindVertexArray(vertex_array);

    //printf("2: %lu, %lu, %lu\n", sizeof(GLfloat), STRIDE, size);
    glGenBuffers(1, &vertex_buffer);
    glBindBuffer(GL_ARRAY_BUFFER, vertex_buffer);
    glBufferData(GL_ARRAY_BUFFER, STRIDE * size, data_array, GL_STATIC_DRAW);

    a_point_position_loc = glGetAttribLocation(shader_program, "aPointPosition");
    a_point_color_loc = glGetAttribLocation(shader_program, "aPointColor");

    glBindVertexArray(0);

    free(data_array);
    return 1;
}

void update() {
}

void render() {
    GLvoid *data_offset = 0;

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    get_identity_matrix(modelview_matrix);

    get_translation_matrix4x4(aux_matrix1, 0.0f, 0.0f, -7.0f);
    mult_matrices(aux_matrix2, modelview_matrix, aux_matrix1, 4);
    memcpy(modelview_matrix, aux_matrix2, 16 * sizeof(float));

    transpose_matrix(aux_matrix1, modelview_matrix, 4);
    transpose_matrix(aux_matrix2, projection_matrix, 4);


    glBindVertexArray(vertex_array);
    glBindBuffer(GL_ARRAY_BUFFER, vertex_buffer);
    glUseProgram(shader_program);

    glUniformMatrix4fv(glGetUniformLocation(shader_program, "uMVMatrix"), 1, GL_FALSE, aux_matrix1);
    glUniformMatrix4fv(glGetUniformLocation(shader_program, "uPMatrix"), 1, GL_FALSE, aux_matrix2);

    glVertexAttribPointer(a_point_position_loc, 3, GL_FLOAT, GL_FALSE, STRIDE, data_offset);
    glEnableVertexAttribArray(a_point_position_loc);

    data_offset += 3 * sizeof(GLfloat);
    glVertexAttribPointer(a_point_color_loc, 3, GL_FLOAT, GL_FALSE, STRIDE, data_offset);
    glEnableVertexAttribArray(a_point_color_loc);

    glDrawArrays(GL_POINTS, 0, size);

    glUseProgram(0);
    glBindVertexArray(0);
}

void cleanup() {
    if (vertex_buffer) {
        glDeleteBuffers(1, &vertex_buffer);
    }
    if (vertex_array) {
        glDeleteVertexArrays(1, &vertex_array);
    }

    if (shader_program && fragment_shader) {
        glDetachShader(shader_program, fragment_shader);
    }
    if (shader_program && vertex_shader) {
        glDetachShader(shader_program, vertex_shader);
    }
    if (fragment_shader) {
        glDeleteShader(fragment_shader);
    }
    if (vertex_shader) {
        glDeleteShader(vertex_shader);
    }
    if (shader_program) {
        glDeleteProgram(shader_program);
    }

    if (gContext) {
        SDL_GL_DeleteContext(gContext);
        gContext = NULL;
    }

    if (gWindow) {
        SDL_DestroyWindow(gWindow);
        gWindow = NULL;
    }

    SDL_Quit();
}



int main(int argc, char* args[]) {
    int quit;
    SDL_Event e;

    if (!init()) {
        printf("Błąd inicjalizacji!\n");
        cleanup();
        return 1;
    }

    quit = 0;
    while (!quit) {
        while (SDL_PollEvent(&e) != 0) {
            if (e.type == SDL_QUIT) {
                quit = 1;
                break;
            }
        }

        update();
        render();

        SDL_GL_SwapWindow(gWindow);
    }

    cleanup();

    return 0;
}



char *text_file_read(char *fname) {
    FILE *fp;
    char *content = NULL;
    int count = 0;

    if (fname != NULL) {
        fp = fopen(fname, "rt");

        if (fp != NULL) {
            fseek(fp, 0, SEEK_END);
            count = ftell(fp);
            rewind(fp);

            if (count > 0) {
                content = (char*)malloc(sizeof(char) * (count+1));
                count = fread(content,sizeof(char),count,fp);
                content[count] = '\0';
            }
            fclose(fp);
        }
    }
    return content;
}


void print_shader_info_log(GLuint obj) {
    int infologLength = 0;
    int charsWritten  = 0;
    char *infoLog;

    glGetShaderiv(obj, GL_INFO_LOG_LENGTH, &infologLength);

    if (infologLength > 0) {
        infoLog = (char*)malloc(infologLength);
        glGetShaderInfoLog(obj, infologLength, &charsWritten, infoLog);
        printf("%s\n",infoLog);
        free(infoLog);
    }
}

void print_program_info_log(GLuint obj) {
    int infologLength = 0;
    int charsWritten  = 0;
    char *infoLog;

    glGetProgramiv(obj, GL_INFO_LOG_LENGTH, &infologLength);

    if (infologLength > 0) {
        infoLog = (char*)malloc(infologLength);
        glGetProgramInfoLog(obj, infologLength, &charsWritten, infoLog);
        printf("%s\n",infoLog);
        free(infoLog);
    }
}


int check_gl_error() {
    GLenum error;

    error = glGetError();
    switch (error) {
        case GL_CONTEXT_LOST:
            printf("GL_CONTEXT_LOST [%d]: %s\n", error, gluErrorString(error));
            return GL_CONTEXT_LOST;
        case GL_INVALID_ENUM:
            printf("GL_INVALID_ENUM [%d]: %s\n", error, gluErrorString(error));
            return GL_INVALID_ENUM;
        case GL_INVALID_VALUE:
            printf("GL_INVALID_VALUE [%d]: %s\n", error, gluErrorString(error));
            return GL_INVALID_VALUE;
        case GL_INVALID_OPERATION:
            printf("GL_INVALID_OPERATION [%d]: %s\n", error, gluErrorString(error));
            return 1;
        case GL_INVALID_FRAMEBUFFER_OPERATION:
            printf("GL_INVALID_FRAMEBUFFER_OPERATION [%d]: %s\n", error, gluErrorString(error));
            return GL_INVALID_FRAMEBUFFER_OPERATION;
        case GL_OUT_OF_MEMORY:
            printf("GL_OUT_OF_MEMORY [%d]: %s\n", error, gluErrorString(error));
            return 1;
        case GL_STACK_OVERFLOW:
            printf("GL_STACK_OVERFLOW [%d]: %s\n", error, gluErrorString(error));
            return GL_STACK_OVERFLOW;
        case GL_STACK_UNDERFLOW:
            printf("GL_STACK_UNDERFLOW [%d]: %s\n", error, gluErrorString(error));
            return GL_STACK_UNDERFLOW;
        case GL_NO_ERROR:
            printf("GL_NO_ERROR [%d]\n", error);
            break;
    }

    return GL_NO_ERROR;
}
