var gl;
var shaderProgram;
var mvMatrix = mat4.create();
var pMatrix = mat4.create();

var pointPositionBuffer;
var pointColorBuffer;


var ratioX;
var ratioY;
var minX = -0.6;
var maxX = 0.2;
var minY = 0.0;
var maxY = 0.8;
var colors = [];
var z_coord = 0.0;

var min_x_coord = -2.9;
var max_x_coord = 2.9;
var min_y_coord = -2.9;
var max_y_coord = 2.9;

var x_step = 0.01;
var y_step = 0.01;

function min(a, b) {
    return (a < b) ? a : b;
}
function max(a, b) {
    return (a > b) ? a : b;
}
// result: [red, green, blue]
function HSV2RGB(hue, sat, val) {
    var i;
    var f, p, q, t;
    
    if (val == 0) {
        //red=0; grn=0; blu=0;
        return [0, 0, 0];
    } else {
        hue /= 60.0;
        i = Math.round(hue);
        f = hue-i;
        p = val * (1 - sat);
        q = val * (1 - (sat * f));
        t = val * (1 - (sat * (1 - f)));
        
        if (i == 0) { return [val, t, p]; }
        else if (i==1) {return [q, val, p]; }
        else if (i==2) {return [p, val, t]; }
        else if (i==3) {return [p, q, val]; }
        else if (i==4) {return [t, p, val]; }
        else if (i==5) {return [val, p, q]; }
    }
}
function initializeColors() {
    var level;
    var r, g, b;
 
    for (var level = 0; level < 120; level++) {
        var color = HSV2RGB(2.1*level, 0.85, 0.6);
        colors[level] = [color[0], color[1], color[2]];
    }
    colors[120] = [0, 0, 0];
}
//calculate squared modus of given complex c
function complexModSq(complex) {
    return complex.real*complex.real + complex.imaginary*complex.imaginary;
}
//function z[0]=0
function f() {
    return {real: 0, imaginary: 0};
}
//function z[n+1] = z[n]^2 + p
function g(z, p) {
    result = {real: 0, imaginary: 0};
    result.real = z.real*z.real - z.imaginary*z.imaginary + p.real;
    result.imaginary = 2*z.real*z.imaginary + p.imaginary;
    return result;
}
//value is inside set in the returned level
function levelSet(p) {
    z = {real: 0, imaginary: 0};
    var iteration;
     
    iteration = 0;
    z = f();
     
    do {
        z = g(z, p);
        iteration++;
    } while (complexModSq(z) < 4 && iteration < 120);
     
    return iteration;
}


function initGL(canvas) {
    try {
        gl = canvas.getContext("webgl");
        gl.viewportWidth = canvas.width;
        gl.viewportHeight = canvas.height;
    } catch (e) {
    }
    if (!gl) {
        alert("Could not initialise WebGL, sorry :-(");
    }
}


function getShader(gl, id) {
    var shaderScript = document.getElementById(id);
    if (!shaderScript) {
        return null;
    }

    var str = "";
    var k = shaderScript.firstChild;
    while (k) {
        if (k.nodeType == 3) {
            str += k.textContent;
        }
        k = k.nextSibling;
    }

    var shader;
    if (shaderScript.type == "x-shader/x-fragment") {
        shader = gl.createShader(gl.FRAGMENT_SHADER);
    } else if (shaderScript.type == "x-shader/x-vertex") {
        shader = gl.createShader(gl.VERTEX_SHADER);
    } else {
        return null;
    }

    gl.shaderSource(shader, str);
    gl.compileShader(shader);

    if (!gl.getShaderParameter(shader, gl.COMPILE_STATUS)) {
        alert(gl.getShaderInfoLog(shader));
        return null;
    }

    return shader;
}


function initShaders() {
    var fragmentShader = getShader(gl, "shader-fs");
    var vertexShader = getShader(gl, "shader-vs");

    shaderProgram = gl.createProgram();
    gl.attachShader(shaderProgram, vertexShader);
    gl.attachShader(shaderProgram, fragmentShader);
    gl.linkProgram(shaderProgram);

    if (!gl.getProgramParameter(shaderProgram, gl.LINK_STATUS)) {
        alert("Could not initialise shaders");
    }

    gl.useProgram(shaderProgram);

    shaderProgram.pointPositionAttribute = gl.getAttribLocation(shaderProgram, "aPointPosition");
    gl.enableVertexAttribArray(shaderProgram.pointPositionAttribute);
    shaderProgram.pointColorAttribute = gl.getAttribLocation(shaderProgram, "aPointColor");
    gl.enableVertexAttribArray(shaderProgram.pointColorAttribute);

    shaderProgram.pMatrixUniform = gl.getUniformLocation(shaderProgram, "uPMatrix");
    shaderProgram.mvMatrixUniform = gl.getUniformLocation(shaderProgram, "uMVMatrix");
}


function initBuffers() {
    var pointsArray = [];
    var colorsArray = [];
    var p = {real: 0, imaginary: 0};
    
    ratioX = (maxX - minX) / 2.0;
    ratioY = (maxY - minY) / 2.0;
    
    initializeColors();
    for (var i = min_x_coord; i < max_x_coord; i += x_step) {
        p.imaginary = i*ratioY + minY;
        for (var j = min_y_coord; j < max_y_coord; j += y_step) {
            p.real = j*ratioX + minX;
            level = levelSet(p);
            pointsArray.push(i);
            pointsArray.push(j);
            pointsArray.push(z_coord);
            
            var c = colors[level];
            colorsArray.push(c[0]);
            colorsArray.push(c[1]);
            colorsArray.push(c[2]);
        }
    }

    //console.log(pointsArray.length);
    //console.log(colorsArray.length);
    console.log(pointsArray);
    console.log(colorsArray);
    
    pointPositionBuffer = gl.createBuffer();
    gl.bindBuffer(gl.ARRAY_BUFFER, pointPositionBuffer);
    gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(pointsArray), gl.STATIC_DRAW);
    pointPositionBuffer.itemSize = 3;
    pointPositionBuffer.numItems = Math.round(pointsArray.length / 3);
    
    pointColorBuffer = gl.createBuffer();
    gl.bindBuffer(gl.ARRAY_BUFFER, pointColorBuffer);
    gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(colorsArray), gl.STATIC_DRAW);
    pointColorBuffer.itemSize = 3;
    pointColorBuffer.numItems = Math.round(colorsArray.length / 3);
}

function setMatrixUniforms() {
    gl.uniformMatrix4fv(shaderProgram.pMatrixUniform, false, pMatrix);
    gl.uniformMatrix4fv(shaderProgram.mvMatrixUniform, false, mvMatrix);
}

function drawScene() {
    gl.viewport(0, 0, gl.viewportWidth, gl.viewportHeight);
    gl.clear(gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT);
    mat4.perspective(45, gl.viewportWidth / gl.viewportHeight, 0.1, 100.0, pMatrix);
    
    //mat4.identity(pMatrix);
    mat4.identity(mvMatrix);
    
    mat4.translate(mvMatrix, [0.0, 0.0, -7.0]);
    
    gl.bindBuffer(gl.ARRAY_BUFFER, pointPositionBuffer);
    gl.vertexAttribPointer(shaderProgram.pointPositionAttribute, pointPositionBuffer.itemSize, gl.FLOAT, false, 0, 0);
    
    gl.bindBuffer(gl.ARRAY_BUFFER, pointColorBuffer);
    gl.vertexAttribPointer(shaderProgram.pointColorAttribute, pointColorBuffer.itemSize, gl.FLOAT, false, 0, 0);
    
    setMatrixUniforms();

    gl.drawArrays(gl.POINT, 0, pointPositionBuffer.numItems);
}



function webGLStart() {
    var canvas = document.getElementById("my-webgl-canvas");
    initGL(canvas);
    initShaders();
    initBuffers();

    gl.clearColor(0.0, 0.0, 0.0, 1.0);
    gl.enable(gl.DEPTH_TEST);

    drawScene();
}
